// Copyright 2019, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Authors: Patrick Brosi <brosi@informatik.uni-freiburg.de>

#include <fstream>
#include <iostream>
#include <set>
#include <sstream>
#include <string>
#include "osmfixer/index/StatIdx.h"
#include "util/geo/BezierCurve.h"
#include "util/log/Log.h"

using osmfixer::Group;
using osmfixer::OsmAttrs;
using osmfixer::StatIdx;
using osmfixer::Station;
using osmfixer::Suggestion;

// _____________________________________________________________________________
void StatIdx::readFromFile(const std::string& path) {
  std::ifstream f(path);

  std::string line;

  // first, parse nodes
  LOG(INFO) << "Parsing nodes...";
  while (std::getline(f, line)) {
    // empty line is separator between blocks
    if (line.size() == 0) break;

    std::stringstream rec(line);
    size_t osmid, origGroup, group;
    double lat, lng;
    OsmAttrs attrs;

    rec >> osmid;
    rec >> lat;
    rec >> lng;
    rec >> origGroup;
    rec >> group;

    std::string temp;
    std::getline(rec, temp, '\t');
    while (std::getline(rec, temp, '\t')) {
      std::string key, val;
      key = temp;
      std::getline(rec, temp, '\t');
      val = temp;
      attrs[key].push_back(val);
    }

    addStation(osmid, lat, lng, origGroup, group, attrs);
  }
  LOG(INFO) << "Done.";

  // second, parse groups
  LOG(INFO) << "Parsing groups... ";
  while (std::getline(f, line)) {
    // empty line is separator between blocks
    if (line.size() == 0) break;

    std::stringstream rec(line);
    size_t osmid;
    OsmAttrs attrs;

    rec >> osmid;

    addGroup(osmid, attrs);
  }

  LOG(INFO) << "Done.";

  // third, parse attr errs
  LOG(INFO) << "Parsing errors... ";
  while (std::getline(f, line)) {
    // empty line is separator between blocks
    if (line.size() == 0) break;

    std::stringstream rec(line);
    size_t id;
    std::string attr, otherAttr;
    double conf;
    size_t otherId;

    rec >> id;

    std::getline(rec, attr, '\t');
    std::getline(rec, attr, '\t');
    std::getline(rec, otherAttr, '\t');

    rec >> conf;
    rec >> otherId;

    _stations[id].attrErrs.push_back({attr, otherAttr, conf, otherId});
  }
  LOG(INFO) << "Done.";

  LOG(INFO) << "Building indices...";
  initGroups();
  initSuggestions();

  initIndex();
  LOG(INFO) << "Done.";
}

// _____________________________________________________________________________
void StatIdx::addStation(size_t osmid, double lat, double lng, size_t origGroup,
                         size_t group, const OsmAttrs& attrs) {
  auto point = util::geo::latLngToWebMerc<double>(lat, lng);

  _stations.emplace_back(
      Station{_stations.size(),
              osmid,
              origGroup,
              group,
              point,
              util::geo::webMercToLatLng<double>(point.getX(), point.getY()),
              attrs,
              {},
              {}});

  // extend bounding box
  _bbox = util::geo::extendBox(point, _bbox);
}

// _____________________________________________________________________________
void StatIdx::addGroup(size_t osmid, const OsmAttrs& attrs) {
  Group g;
  g.id = _groups.size();
  g.osmid = osmid;
  g.attrs = attrs;
  _groups.emplace_back(g);
}

// _____________________________________________________________________________
void StatIdx::initGroups() {
  for (size_t i = 0; i < _stations.size(); i++) {
    // this should be ensured by the input file
    assert(_stations[i].origGroup < _groups.size());
    _groups[_stations[i].origGroup].polyStations.push_back(i);

    _groups[_stations[i].group].stations.push_back(i);

    if (_stations[i].group != _stations[i].origGroup) {
      _groups[_stations[i].origGroup].stations.push_back(i);
    }

    assert(_stations[i].group < _groups.size());
    if (_stations[i].group != _stations[i].origGroup &&
        _groups[_stations[i].group].osmid == 1) {
      // only add non-orig stations to polygons of new (non-osm) groups
      _groups[_stations[i].group].polyStations.push_back(i);
    }
  }

  // build hull polygon
  for (size_t i = 0; i < _groups.size(); i++) {
    util::geo::MultiPoint<double> mp;
    for (size_t stid : _groups[i].polyStations) {
      double rad = 11.0;
      int n = 20;
      for (int i = 0; i < n; i++) {
        double x = rad * cos((2.0 * M_PI / static_cast<double>(n)) *
                             static_cast<double>(i));
        double y = rad * sin((2.0 * M_PI / static_cast<double>(n)) *
                             static_cast<double>(i));

        mp.push_back(util::geo::DPoint(_stations[stid].pos.getX() + x,
                                       _stations[stid].pos.getY() + y));
      }
    }
    _groups[i].poly = util::geo::convexHull(mp);
  }
}

// _____________________________________________________________________________
void StatIdx::initIndex() {
  double gSize = 10000;
  _sgrid = util::geo::Grid<size_t, util::geo::Point, double>(gSize, gSize,
                                                             _bbox, false);
  _ggrid = util::geo::Grid<size_t, util::geo::Polygon, double>(gSize, gSize,
                                                               _bbox, false);
  _suggrid = util::geo::Grid<size_t, util::geo::Line, double>(gSize, gSize,
                                                              _bbox, false);

  for (size_t i = 0; i < _stations.size(); i++) _sgrid.add(_stations[i].pos, i);
  for (size_t i = 0; i < _groups.size(); i++) _ggrid.add(_groups[i].poly, i);
  for (size_t i = 0; i < _suggestions.size(); i++)
    _suggrid.add(_suggestions[i].arrow, i);

  _heatGridsOk.resize(10);
  _heatGridsSugg.resize(10);
  _heatGridsErr.resize(10);

  for (size_t i = 0; i < 10; i++) {
    _heatGridsOk[i] = (util::geo::Grid<osmfixer::Cluster, util::geo::Point, double>(
        5000, 5000, _bbox, false));
    _heatGridsSugg[i] = (util::geo::Grid<osmfixer::Cluster, util::geo::Point, double>(
        5000, 5000, _bbox, false));
    _heatGridsErr[i] = (util::geo::Grid<osmfixer::Cluster, util::geo::Point, double>(
        5000, 5000, _bbox, false));
    int step = 2000 * (i + 1);

    for (size_t x = 1; x * step < (_sgrid.getXWidth() * gSize); x++) {
      for (size_t y = 1; y * step < (_sgrid.getYHeight() * gSize); y++) {
        std::set<size_t> tmp;
        auto reqBox = util::geo::DBox({_bbox.getLowerLeft().getX() + (x - 1) * step, _bbox.getLowerLeft().getY() + (y - 1) * step}, {_bbox.getLowerLeft().getX() + x * step, _bbox.getLowerLeft().getY() + y * step});
        _sgrid.get(reqBox, &tmp);

        size_t countOk = 0;
        double avgXOk = 0;
        double avgYOk = 0;

        size_t countSugg = 0;
        double avgXSugg = 0;
        double avgYSugg = 0;

        size_t countErr = 0;
        double avgXErr = 0;
        double avgYErr = 0;

        for (auto j : tmp) {
          const auto& stat = _stations[j];
          if (stat.suggestions.size() == 0 && stat.attrErrs.size() == 0) {
            if (util::geo::contains(stat.pos, reqBox)) {
              countOk++;
              avgXOk += stat.pos.getX();
              avgYOk += stat.pos.getY();
            }
          } else if (stat.suggestions.size() > 0 && stat.attrErrs.size() == 0) {
            if (util::geo::contains(stat.pos, reqBox)) {
              countSugg++;
              avgXSugg += stat.pos.getX();
              avgYSugg += stat.pos.getY();
            }
          } else if (stat.attrErrs.size() > 0) {
            if (util::geo::contains(stat.pos, reqBox)) {
              countErr++;
              avgXErr += stat.pos.getX();
              avgYErr += stat.pos.getY();
            }
          }
        }

        if (countOk != 0) {
          avgXOk /= countOk;
          avgYOk /= countOk;
          auto ll = util::geo::webMercToLatLng<double>(avgXOk, avgYOk);

          _heatGridsOk[i].add({avgXOk, avgYOk}, Cluster{{avgXOk, avgYOk}, ll, countOk});
        }

        if (countSugg != 0) {
          avgXSugg /= countSugg;
          avgYSugg /= countSugg;
          auto ll = util::geo::webMercToLatLng<double>(avgXSugg, avgYSugg);

          _heatGridsSugg[i].add({avgXSugg, avgYSugg}, Cluster{{avgXSugg, avgYSugg}, ll, countSugg});
        }

        if (countErr != 0) {
          avgXErr /= countErr;
          avgYErr /= countErr;
          auto ll = util::geo::webMercToLatLng<double>(avgXErr, avgYErr);

          _heatGridsErr[i].add({avgXErr, avgYErr}, Cluster{{avgXErr, avgYErr}, ll, countErr});
        }
      }
    }
  }
}

// _____________________________________________________________________________
std::vector<const Suggestion*> StatIdx::getSuggestions(
    const util::geo::DBox bbox) const {
  std::vector<const Suggestion*> ret;
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  std::set<size_t> tmp;
  auto reqBox = util::geo::DBox(ll, ur);
  _suggrid.get(reqBox, &tmp);

  for (auto i : tmp) {
    if (util::geo::intersects(_suggestions[i].arrow, reqBox))
      ret.push_back(&_suggestions[i]);
  }

  return ret;
}

// _____________________________________________________________________________
std::vector<const Group*> StatIdx::getGroups(const util::geo::DBox bbox) const {
  std::vector<const Group*> ret;
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  std::set<size_t> tmp;
  auto reqBox = util::geo::DBox(ll, ur);
  _ggrid.get(reqBox, &tmp);

  for (auto i : tmp) {
    // if (util::geo::intersects(_groups[i].poly, reqBox))
      ret.push_back(&_groups[i]);
  }

  return ret;
}

// _____________________________________________________________________________
std::vector<const Station*> StatIdx::getStations(
    const util::geo::DBox bbox) const {
  std::vector<const Station*> ret;
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  std::set<size_t> tmp;
  auto reqBox = util::geo::DBox(ll, ur);
  _sgrid.get(reqBox, &tmp);

  for (auto i : tmp) {
    if (util::geo::contains(_stations[i].pos, reqBox))
      ret.push_back(&_stations[i]);
  }

  return ret;
}

// _____________________________________________________________________________
const Station* StatIdx::getStation(size_t id) const {
  if (id >= _stations.size()) return 0;
  return &_stations[id];
}

// _____________________________________________________________________________
std::vector<osmfixer::Cluster> StatIdx::getHeatGridErr(const util::geo::DBox bbox,
                                                    size_t z) const {
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  auto reqBox = util::geo::DBox(ll, ur);

  z = fmin(9, z);
  z = fmax(0, z);

  const auto& grid = _heatGridsErr[9 - z];
  std::vector<osmfixer::Cluster> ret;

  std::set<osmfixer::Cluster> tmp;
  grid.get(reqBox, &tmp);

  for (const auto& i : tmp) {
    if (util::geo::contains(i.pos, reqBox)) ret.push_back(i);
  }

  return ret;
}

// _____________________________________________________________________________
std::vector<osmfixer::Cluster> StatIdx::getHeatGridOk(const util::geo::DBox bbox,
                                                    size_t z) const {
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  auto reqBox = util::geo::DBox(ll, ur);

  z = fmin(9, z);
  z = fmax(0, z);

  const auto& grid = _heatGridsOk[9 - z];
  std::vector<osmfixer::Cluster> ret;

  std::set<osmfixer::Cluster> tmp;
  grid.get(reqBox, &tmp);

  for (const auto& i : tmp) {
    if (util::geo::contains(i.pos, reqBox)) ret.push_back(i);
  }

  return ret;
}

// _____________________________________________________________________________
std::vector<osmfixer::Cluster> StatIdx::getHeatGridSugg(const util::geo::DBox bbox,
                                                    size_t z) const {
  auto ll = util::geo::latLngToWebMerc<double>(bbox.getLowerLeft().getX(),
                                               bbox.getLowerLeft().getY());
  auto ur = util::geo::latLngToWebMerc<double>(bbox.getUpperRight().getX(),
                                               bbox.getUpperRight().getY());

  auto reqBox = util::geo::DBox(ll, ur);

  z = fmin(9, z);
  z = fmax(0, z);

  const auto& grid = _heatGridsSugg[9 - z];
  std::vector<osmfixer::Cluster> ret;

  std::set<osmfixer::Cluster> tmp;
  grid.get(reqBox, &tmp);

  for (const auto& i : tmp) {
    if (util::geo::contains(i.pos, reqBox)) ret.push_back(i);
  }

  return ret;
}
// _____________________________________________________________________________
void StatIdx::initSuggestions() {
  for (size_t i = 0; i < _stations.size(); i++) {
    auto& stat = _stations[i];
    if (stat.group != stat.origGroup || getGroup(stat.group)->osmid == 1) {
      Suggestion sug;
      sug.station = i;
      sug.arrow = util::geo::DLine{stat.pos, stat.pos};
      if (getGroup(stat.origGroup)->osmid < 2) {
        if (getGroup(stat.group)->osmid == 1) {
          // move orphan into new group
          sug.type = 1;
          sug.target_gid = stat.group;

          // sug.arrow = getGroupArrow(i, stat.group);

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);

        } else if (getGroup(stat.group)->osmid > 1) {
          // move orphan into existing group
          sug.type = 2;
          sug.target_gid = stat.group;
          sug.target_osm_rel_id = getGroup(stat.group)->osmid;

          auto b = util::geo::centroid(getGroup(stat.group)->poly);
          sug.arrow = getGroupArrow(i, b);

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        }
      } else {
        if (getGroup(stat.group)->osmid == 1 &&
            getGroup(stat.group)->stations.size() > 1) {
          // move station from relation into new group
          sug.type = 3;
          sug.orig_gid = stat.origGroup;
          sug.orig_osm_rel_id = getGroup(stat.origGroup)->osmid;
          sug.target_gid = stat.group;

          auto b = util::geo::centroid(getGroup(stat.group)->poly);
          sug.arrow = getGroupArrow(i, b);

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        } else if (getGroup(stat.group)->osmid > 1) {
          // move station from relation into existing group
          sug.type = 4;
          sug.orig_gid = stat.origGroup;
          sug.orig_osm_rel_id = getGroup(stat.origGroup)->osmid;
          sug.target_gid = stat.group;
          sug.target_osm_rel_id = getGroup(stat.group)->osmid;

          auto b = util::geo::centroid(getGroup(stat.group)->poly);
          sug.arrow = getGroupArrow(i, b);

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        } else {
          // move station out of relation
          sug.type = 5;
          sug.orig_gid = stat.origGroup;
          sug.target_gid = stat.group;
          sug.target_osm_rel_id = getGroup(stat.group)->osmid;

          auto b = stat.pos;
          b.setX(b.getX() + 50);
          b.setY(b.getY() + 50);

          sug.arrow = getGroupArrow(i, b);

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        }
      }

      for (auto attrErr : stat.attrErrs) {
        const auto otherStat = getStation(attrErr.otherId);
        if (otherStat->osmid == stat.osmid) {
          // fix attributes
          sug.type = 6;
          sug.orig_gid = 0;
          sug.orig_osm_rel_id = 0;
          sug.target_gid = 0;
          sug.target_osm_rel_id = 0;
          sug.attrErrName = attrErr.attr;

          _suggestions.push_back(sug);
          stat.suggestions.push_back(_suggestions.size() - 1);
        }
      }
    }
  }
}

// _____________________________________________________________________________
util::geo::DLine StatIdx::getGroupArrow(size_t stat, const util::geo::DPoint& b) const {
  auto a = getStation(stat)->pos;

  auto pl = util::geo::PolyLine<double>(a, b);
  auto bb = pl.getPointAtDist(fmax(pl.getLength() / 2, pl.getLength() - 5)).p;

  if (util::geo::dist(a, bb) < 1) {
    bb = a;
    bb.setX(bb.getX() + 3);
  }

  util::geo::Line<double> line;

  if (util::geo::dist(a, bb) < 20) {
    line = {a, bb};
  } else {

    int side = rand() % 2;
    if (!side) side = -1;

    auto rot1 = util::geo::rotate(util::geo::DLine{a, bb}, 45 * side, a);
    auto rot2 = util::geo::rotate(util::geo::DLine{a, bb}, -20 * side, bb);
    auto rot3 = util::geo::rotate(util::geo::DLine{a, bb}, -120 * side, bb);

    auto i =
        util::geo::intersection(util::geo::LineSegment<double>{rot1[0], rot1[1]},
                                util::geo::LineSegment<double>{rot2[0], rot2[1]});

    line =
        util::geo::BezierCurve<double>(a, i, rot3[0], bb).render(5).getLine();
  }

  // arrowhead
  auto pl2 = util::geo::PolyLine<double>(line);
  util::geo::LineSegment<double> headSeg(
      pl2.getPointAtDist(pl2.getLength() - 5).p, line.back());
  auto arrHeadLeft = util::geo::rotate(headSeg, -45, line.back()).first;
  auto arrHeadRight = util::geo::rotate(headSeg, 45, line.back()).first;

  auto tmp = line.back();

  line.push_back(arrHeadLeft);
  line.push_back(tmp);
  line.push_back(arrHeadRight);

  return line;
}

// _____________________________________________________________________________
const Group* StatIdx::getGroup(size_t id) const {
  if (id >= _groups.size()) return 0;
  return &_groups[id];
}

// _____________________________________________________________________________
const Suggestion* StatIdx::getSuggestion(size_t id) const {
  if (id >= _suggestions.size()) return 0;
  return &_suggestions[id];
}
