// Copyright 2019, University of Freiburg,
// Chair of Algorithms and Data Structures.
// Authors: Patrick Brosi <brosi@informatik.uni-freiburg.de>

#ifndef OSMFIXER_INDEX_STATIDX_H_
#define OSMFIXER_INDEX_STATIDX_H_

#include <map>
#include <string>
#include <vector>
#include "util/geo/Geo.h"
#include "util/geo/Grid.h"

namespace osmfixer {

typedef std::map<std::string, std::vector<std::string>> OsmAttrs;

struct AttrErr {
  std::string attr;
  std::string otherAttr;
  double conf;
  size_t otherId;
};

struct Suggestion {
  uint16_t type;
  size_t station;
  size_t target_gid, orig_gid, target_osm_rel_id, orig_osm_rel_id;
  std::string attrErrName;
  util::geo::DLine arrow;
};

struct Station {
  size_t id;
  size_t osmid;
  size_t origGroup, group;
  util::geo::DPoint pos;
  util::geo::DPoint latLng;
  OsmAttrs attrs;
  std::vector<AttrErr> attrErrs;
  std::vector<size_t> suggestions;
};

struct Group {
  size_t id;
  size_t osmid;
  util::geo::DPolygon poly;
  OsmAttrs attrs;
  std::vector<size_t> stations;
  std::vector<size_t> polyStations;
};

struct Cluster {
  util::geo::DPoint pos;
  util::geo::DPoint latLng;
  size_t size;
  bool operator<(const Cluster& other) const { return this < &other; }
};

class StatIdx {
 public:
  StatIdx() {}

  void readFromFile(const std::string& path);

  std::vector<const Station*> getStations(const util::geo::DBox bbox) const;
  std::vector<const Group*> getGroups(const util::geo::DBox bbox) const;
  std::vector<const Suggestion*> getSuggestions(
      const util::geo::DBox bbox) const;

  const Station* getStation(size_t id) const;
  const Group* getGroup(size_t id) const;
  const Suggestion* getSuggestion(size_t id) const;

  std::vector<Cluster> getHeatGridOk(const util::geo::DBox bbox,
      size_t z) const;

  std::vector<Cluster> getHeatGridSugg(const util::geo::DBox bbox,
      size_t z) const;

  std::vector<Cluster> getHeatGridErr(const util::geo::DBox bbox,
      size_t z) const;

 private:
  void addStation(size_t id, double lat, double lng, size_t origGroup,
                  size_t group, const OsmAttrs& attrs);
  void addGroup(size_t id, const OsmAttrs& attrs);

  void initIndex();
  void initGroups();
  void initSuggestions();

  util::geo::DLine getGroupArrow(size_t stat, const util::geo::DPoint& p) const;

  std::vector<Station> _stations;
  std::vector<Group> _groups;
  std::vector<Suggestion> _suggestions;
  util::geo::DBox _bbox;

  util::geo::Grid<size_t, util::geo::Point, double> _sgrid;
  util::geo::Grid<size_t, util::geo::Polygon, double> _ggrid;
  util::geo::Grid<size_t, util::geo::Line, double> _suggrid;

  std::vector<util::geo::Grid<Cluster, util::geo::Point, double>> _heatGridsOk;
  std::vector<util::geo::Grid<Cluster, util::geo::Point, double>> _heatGridsSugg;
  std::vector<util::geo::Grid<Cluster, util::geo::Point, double>> _heatGridsErr;
};
}  // namespace osmfixer

#endif  // OSMFIXER_INDEX_STATIDX_H_
